/*
	dcc-kiss
	
	Simple serial to DCC converter.

	Copyright Jeroen Vreeken (jeroen@vreeken.net), 2010, 2011, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <stdbool.h>

#define F_CPU 8000000


#define SLIP_END 192
#define SLIP_ESC 219
#define SLIP_ESC_END 220
#define SLIP_ESC_ESC 221


static bool rail_power = true;

enum mod_state {
	MOD_ST_IDLE,
	MOD_ST_CUTOUT,
	MOD_ST_PACKET,
};

#define MOD_BUFFER_LEN	512

#define MOD_SYM_DISABLE	0
#define MOD_SYM_ENABLE  4
#define MOD_SYM_POS	(8 | MOD_SYM_ENABLE)
#define MOD_SYM_NEG	(16 | MOD_SYM_ENABLE)
#define MOD_SYM_CUTOUT	MOD_SYM_ENABLE

#define MOD_PREAMBLE_BITS	14

#define MOD_LEN_ONE		4
#define MOD_LEN_ZERO		8
#define MOD_LEN_CUTOUT_BLANK	(4*4-1)
#define MOD_LEN_PREAMBLE	(MOD_PREAMBLE_BITS*MOD_LEN_ONE)
#define MOD_LEN_BYTE		(9*MOD_LEN_ZERO) /* worst case: all zeros */
#define MOD_LEN_IDLE		(MOD_LEN_PREAMBLE + MOD_LEN_BYTE * 3)
#define MOD_LEN_CUTOUT		(MOD_LEN_ONE * 2 + MOD_LEN_ONE * 4)
#define MOD_LEN_PACKET(bytes)	(MOD_LEN_PREAMBLE + MOD_LEN_BYTE * (bytes))

static enum mod_state mod_state = MOD_ST_IDLE;
static unsigned char mod_buffer[MOD_BUFFER_LEN];
static volatile int mod_rd = 0;
static int mod_wr = 0;

#define SLIP_BUFFER_SIZE 16

static unsigned char slip_buffer[SLIP_BUFFER_SIZE];
static int slip_pos = 0;
static bool slip_first = true;

enum slip_state {
	SLIP_ST_READING,
	SLIP_ST_HANDLING,
};

static enum slip_state slip_state = SLIP_ST_READING;

#define PACKET_HEADER_LEN	1
#define PACKET_HEADER_DCC	0
#define PACKET_HEADER_POWER_ON	1
#define PACKET_HEADER_POWER_OFF	2
#define PACKET_HEADER_CUTOUT	3

#define ACK_PACKET	0
#define ACK_POWER_ON	1
#define ACK_POWER_OFF	2
#define ACK_CUTOUT	3
#define NACK	0xff

static void slip_tx(unsigned char ack);


static void mod_add_symbol(unsigned char symbol)
{
	mod_buffer[mod_wr] = symbol;
	mod_wr++;
	mod_wr &= MOD_BUFFER_LEN-1;
}

static void mod_add_bit(bool bit)
{
	mod_add_symbol(MOD_SYM_NEG);
	mod_add_symbol(MOD_SYM_NEG);
	if (!bit) {
		mod_add_symbol(MOD_SYM_NEG);
		mod_add_symbol(MOD_SYM_NEG);
	}
	mod_add_symbol(MOD_SYM_POS);
	mod_add_symbol(MOD_SYM_POS);
	if (!bit) {
		mod_add_symbol(MOD_SYM_POS);
		mod_add_symbol(MOD_SYM_POS);
	}
}

static void mod_add_preamble(void)
{
	int i;
	
	for (i = MOD_PREAMBLE_BITS; i; i--) {
		mod_add_bit(1);
	} 
}

static void mod_add_byte(unsigned char byte)
{
	int i;
	
	mod_add_bit(0);
	for (i = 128; i; i /= 2) {
		mod_add_bit(byte & i);
	}
}

static void mod(void) {
	unsigned int mod_room = mod_rd - mod_wr;
	int i;

	mod_room &= MOD_BUFFER_LEN-1;

	if (!mod_room)
		return;
	
	if (!rail_power) {
		mod_add_symbol(MOD_SYM_DISABLE);
		mod_state = MOD_ST_IDLE;
		return;
	}

	if (mod_state == MOD_ST_IDLE && mod_room >= MOD_LEN_IDLE) {
		mod_add_preamble();
		mod_add_byte(0xff);
		mod_add_byte(0x00);
		mod_add_byte(0xff);
	}
	if (mod_state == MOD_ST_CUTOUT && mod_room >= MOD_LEN_CUTOUT) {
		mod_add_bit(1);
		mod_add_symbol(MOD_SYM_NEG);
		for (i = 0; i < MOD_LEN_CUTOUT_BLANK; i++)
			mod_add_symbol(MOD_SYM_CUTOUT);
		mod_state = MOD_ST_IDLE;
		slip_tx(ACK_CUTOUT);
	}
	if (mod_state == MOD_ST_PACKET && 
	    mod_room >= MOD_LEN_PACKET(slip_pos - 1)) {
		mod_add_preamble();
		for (i = 1; i < slip_pos; i++)
			mod_add_byte(slip_buffer[i]);
		mod_state = MOD_ST_IDLE;
		slip_tx(ACK_PACKET);
	}
}

ISR(TIMER1_COMPA_vect)
{
	PORTB =  mod_buffer[mod_rd];
	mod_rd++;
	mod_rd &= MOD_BUFFER_LEN-1;
}


#define UART_RX_BUFFER_SIZE 64

static unsigned char uart_rx_buffer[UART_RX_BUFFER_SIZE];
static int uart_rx_wr = 0;
static int uart_rx_rd = 0;

#define UART_TX_BUFFER_SIZE 16

static unsigned char uart_tx_buffer[UART_TX_BUFFER_SIZE];
static int uart_tx_wr = 0;
static int uart_tx_rd = 0;

static void uart_tx_byte(unsigned char val)
{
	uart_tx_buffer[uart_tx_wr] = val;
	uart_tx_wr++;
	uart_tx_wr &= UART_TX_BUFFER_SIZE-1;
}

static void uart(void)
{
	unsigned char newbyte;
	
	if (UCSR0A & (1<<RXC0)) {
		newbyte = UDR0;
		uart_rx_buffer[uart_rx_wr] = newbyte;
		uart_rx_wr++;
		uart_rx_wr &= UART_RX_BUFFER_SIZE-1;
	}
	if ((UCSR0A & (1<<UDRE0)) && uart_tx_wr != uart_tx_rd) {
		UDR0 = uart_tx_buffer[uart_tx_rd];
		uart_tx_rd++;
		uart_tx_rd &= UART_TX_BUFFER_SIZE-1;
	}
}

static void slip_tx(unsigned char ack)
{
	uart_tx_byte(SLIP_END);
	switch(ack) {
		case SLIP_END:
			uart_tx_byte(SLIP_ESC);
			uart_tx_byte(SLIP_ESC_END);
			break;
		case SLIP_ESC:
			uart_tx_byte(SLIP_ESC);
			uart_tx_byte(SLIP_ESC_ESC);
			break;
		default:
			uart_tx_byte(ack);
			break;
	}
	uart_tx_byte(SLIP_END);
}

static void slip_rx(void)
{
	int length = uart_rx_wr - uart_rx_rd;
	length &= UART_RX_BUFFER_SIZE-1;
	int i;
	unsigned char uart_c = uart_rx_buffer[uart_rx_rd];

	if (slip_state == SLIP_ST_HANDLING) {
		return;
	}

	if (length) switch (uart_c) {
		case SLIP_END:
			if (!slip_first) {
				if (slip_pos) {
					slip_state = SLIP_ST_HANDLING;
				}
			} else {
				slip_first = false;
			}
			uart_rx_rd++;
			break;
		case SLIP_ESC: {
			unsigned char c;
			
			if (length < 2)
				break;
			i = uart_rx_rd + 1;
			i &= UART_RX_BUFFER_SIZE-1;
			if (uart_rx_buffer[i] == SLIP_ESC_END) {
				c = SLIP_END;
			}
			if (uart_rx_buffer[i] == SLIP_ESC_ESC) {
				c = SLIP_ESC;
			}
			slip_buffer[slip_pos] = c;
			slip_pos++;
			uart_rx_rd += 2;
			break;
		}
		default:
			slip_buffer[slip_pos] = uart_c;
			slip_pos++;
			uart_rx_rd++;
			break;
	}
	uart_rx_rd &= UART_RX_BUFFER_SIZE-1;
}

enum packet_state {
	PACKET_ST_HEADER,
	PACKET_ST_MOD,
};

static enum packet_state packet_state = PACKET_ST_HEADER;

static void handle_packet(void)
{
	if (slip_state != SLIP_ST_HANDLING) {
		return;
	}
	
	if (slip_pos >= PACKET_HEADER_LEN) {
		if (packet_state == PACKET_ST_HEADER) {
			switch (slip_buffer[0]) {
				case PACKET_HEADER_POWER_OFF:
					rail_power = false;
					slip_tx(ACK_POWER_OFF);
					break;
				case PACKET_HEADER_POWER_ON:
					rail_power = true;
					slip_tx(ACK_POWER_ON);
					break;
				case PACKET_HEADER_CUTOUT:
					if (rail_power) {
						packet_state = PACKET_ST_MOD;
						mod_state = MOD_ST_CUTOUT;
					} else {
						slip_tx(NACK);
					}
					break;
				case PACKET_HEADER_DCC:
					if (rail_power) {
						packet_state = PACKET_ST_MOD;
						mod_state = MOD_ST_PACKET;
					} else {
						slip_tx(NACK);
					}
					break;
				default:
					slip_tx(NACK);
					break;
			}
		}
	}
	
	/* wait untill mod returns to idle state */
	if (packet_state == PACKET_ST_MOD) {
		if (mod_state == MOD_ST_IDLE)
			packet_state = PACKET_ST_HEADER;
	}
	
	/* Are we done? then go wait for next packet */
	if (packet_state == PACKET_ST_HEADER) {
		slip_pos = 0;
		slip_state = SLIP_ST_READING;
	}
}

static void uart_9600(void)
{
#define BAUD 9600
#include <util/setbaud.h>
	UBRR0H = UBRRH_VALUE;
	UBRR0L = UBRRL_VALUE;
	#if USE_2X
	UCSR0A |= (1 << U2X0);
	#else
	UCSR0A &= ~(1 << U2X0);
	#endif
	UCSR0B = (1<<RXEN0) | (1<<TXEN0);
	UCSR0C = 3<<UCSZ00;
}

static void counter(void)
{
	/* Interval: 29us (quarter of a 'one') */
	TCCR1A = 0;
	TCCR1B = (1<<WGM12) | (1<<CS10);
	OCR1AH = 0x00;
	OCR1AL = 0xe8;
	TIMSK1 = 1<<OCIE1A;
}

int main(void)
{
	int i;
	wdt_reset();
	
	/* port B direction */
	DDRB = 0xff;
	
	uart_9600();
	counter();

	for (i = 0; i < MOD_BUFFER_LEN; i++)
		mod_buffer[i] = MOD_SYM_DISABLE;
	
	SREG |= 1<<SREG_I;
	
	while (1){
		wdt_reset();
		uart();
		slip_rx();
		handle_packet();
		mod();
	}
	
	return 0;
}
